using UnityEngine;
using System.Collections;
using System.Linq;
using System.Collections.Generic;
using System.Collections.ObjectModel;




public class GameState {
	
	public List<Entity> entities;	
	public Dictionary<Position,List<Entity>> entitybuckets;
	public Dictionary<System.Type,List<Entity>> entitytypebuckets;
	public List<BeamDat> beams = new List<BeamDat>();
	public bool day;
	
	public bool WasOccupied(Position p)
	{
		foreach(Entity e in entities)
		{
			if (e.blockslight && e.lastmovedir!= Position.Direction.MAX)
			{
				if (e.pos-e.lastmovedir==p)
					return true;
			}
		}
		return false;
	}
	public List<Entity>At(Position p)
	{
		List<Entity> result;
		if (entitybuckets.TryGetValue(p,out result)==false)
		{
			result = new List<Entity>();
			entitybuckets.Add(new Position(p),result);
		}
		return result;
	}
	
	//only call form postdeserialize
	void PopulateBeamList()
	{
		beams = Get<Whale>().SelectMany( w => ((Whale)w).beams ).ToList();
		beams.AddRange(Get<StaticWhale>().SelectMany( w => ((Whale)w).beams ));
	}
	
	public void Rotate()
	{
		entities.ForEach( e => 
		{
			e.pos = e.pos.Rotated();
			e.facing = e.facing.Rotate();
		}
		);
		PostDeserialize();
	}
	public void HFlip()
	{
		entities.ForEach( e => 
		{
			e.pos = e.pos.HFlipped();
			e.facing = e.facing.HFlip();
		}
		);
		PostDeserialize();
	}
	public void VFlip()
	{
		entities.ForEach( e => 
		{
			e.pos = e.pos.VFlipped();
			e.facing = e.facing.VFlip();
		}
		);
		PostDeserialize();
	}
	
	private void PostDeserialize()
	{
		RebuildBuckets();
		PopulateBeamList();
		PostMoveCalcs();
		player = (Player)Get<Player>().First();
		SanityCheck();
	}
	
	void ZeroLastMovedDirs()
	{
		foreach(Entity e in entities)
		{
			e.lastmovedir = Position.Direction.MAX;
			e.animdelay = 0.0f;
		}
	}
	
	private void PreMoveCalcs()
	{
		ZeroLastMovedDirs();
		Level.ThisTurnLength=0.0f;
	}
				
	public void PostMoveCalcs()
	{
		CalcBeams();	
		Level.ThisTurnLength+=Level.TurnLength;
		CheckDaytime();
		CheckWin();
	}
	
	//automatically calculated, no need to do fancy stuff
	public bool won=false;
	
	public void CheckWin()
	{
		if (day==false)
		{
			won = false;
			return;
		}
		
		List<Entity> et = Get<EndTile>();
		if (player!=null && et!=null && et.Any( e => e.pos == player.pos ))
		{
			won = true;
		}
		
	}
	
	public void CheckDaytime()
	{
		if (player==null)
			return;
		
		if (day)
		{
			if (beams.Any( b=> b.owner.pos + b.length*b.dir.ToPosition() == player.pos))
			{
				day=false;
			}
		}
		else
		{			
			if (Get<StartTile>().Any( st => st.pos == player.pos))
			{
				day=true;
			}
		}
	}
		
	private void RebuildBuckets()
	{
		//Debug.Log("rebuilding buckets");
		entitybuckets = new Dictionary<Position, List<Entity>>();
		
		
		entitytypebuckets = new Dictionary<System.Type, List<Entity>>();
		entitytypebuckets.Add(typeof(Block),new List<Entity>());
		entitytypebuckets.Add(typeof(Player),new List<Entity>());
		entitytypebuckets.Add(typeof(Wall),new List<Entity>());
		entitytypebuckets.Add(typeof(Whale),new List<Entity>());
		entitytypebuckets.Add(typeof(StartTile),new List<Entity>());
		entitytypebuckets.Add(typeof(EndTile),new List<Entity>());
		entitytypebuckets.Add(typeof(StaticWhale),new List<Entity>());
		
		//entitybuckets = new Dictionary<Position, List<Entity>>();
		foreach(Entity e in entities)
		{
			//entity bucket
			List<Entity> bucket;
			if (entitybuckets.TryGetValue(e.pos, out bucket)==false)
			{
				bucket = new List<Entity>();
				entitybuckets.Add(new Position(e.pos),bucket);
			}
			bucket.Add(e);
			
			//entity type bucket			
			entitytypebuckets[e.GetType()].Add(e);
		}				
	}
	
	public void AddEntity(Entity e, bool temp)
	{
		if (e.GetType()==typeof(Player))
		{
			player=(Player)e;
		}		
		
		if (temp==false)
		{
			
			if (e.beamsource)
			{
				Whale w = (Whale)e;
				beams.AddRange(w.beams);
			}
			
			entities.Add(e);
			entitytypebuckets[e.GetType()].Add(e);	
		}		
		
		List<Entity> bucket;
		if (entitybuckets.TryGetValue(e.pos, out bucket)==false)
		{
			bucket = new List<Entity>();
			entitybuckets.Add(new Position(e.pos),bucket);
		}		
		bucket.Add(e);
		
			
	}
	
	public void RemoveEntity(Entity e,bool temp)
	{			
		List<Entity> bucket = entitybuckets[e.pos];
		bucket.Remove(e);
				
		
		if (temp==false)
		{
			System.Type etype = e.GetType();
			entitytypebuckets[etype].Remove(e);
			
			entities.Remove(e);					
			
			if (e.beamsource)
			{
				Whale w = (Whale)e;//staticwhales can still be cast to whales
				beams.RemoveAll( b => b.owner == w);
			}
			
			if (bucket.Count==0)
			{
				entitybuckets.Remove(e.pos);
			}
		}				
	}
	
	public void SanityCheck()
	{
		foreach(var k in entitybuckets.Keys)
		{
			foreach (var e in entitybuckets[k])
			{
				if (e.pos!=k)
				{
					Debug.LogError("wrong bucket " +e+ " - "+k);
				}
			}
		}
		foreach( var e in entities)
		{
			if (entitybuckets[e.pos].Contains(e)==false)
			{
				Debug.Log("couldn't find " + e);
			}
		}
		
		foreach(BeamDat bd in beams)
		{
			if ( At(bd.owner.pos).Any( w => bd.owner == w ) ==false )
			{
				Debug.Log("beam positioned at " + bd.owner.pos + " has no owner at that point.");
			}
		}		
	}
	
	public void CalcBeams()
	{
		Get<Whale>().ForEach( w => ((Whale)w).CalcBeams());	
		Get<StaticWhale>().ForEach( w => ((StaticWhale)w).CalcBeams());		
	}
	
	
	public bool BlocksLight(Position p, ref Entity blocker)
	{
		var ents = At(p);
		foreach(Entity e in ents)
		{
			if (e.blockslight)
			{
				blocker=e;
				return true;
			}
		}
		
		return false;
	}
	
	public List<Entity> Get<T>() where T : Entity
	{
		return entitytypebuckets[typeof(T)];// (IEnumerable<T>)entities.Where( e => e.GetType()==typeof(T) ).Select(e => (T)e);
	}		
	
	//ACCESSORS END
	
	public GameState()
	{
		entities = new List<Entity>();
		RebuildBuckets();
	}	
		
	private Player player;
	public Player Player()
	{
		return player;//(Player)entities.Find( e => e.GetType()==typeof(Player));
	}
	
	
	public void BroadcastMessage(Entity e, Position.Direction d, Entity.MoverType othertype, float delay)
	{
		entities.ForEach( ent => ent.PreMoveMessage(e,d,othertype,delay) );
	}
	
	private List<List<Position>> positionbaks = new List<List<Position>>();
	public void BackupPositions()
	{
		List<Position> positions = new List<Position>(entities.Count);
		for (int i=0;i<entities.Count;i++)
		{
			Entity e = entities[i];
			Position p = new Position(e.pos);
			positions.Add(p);
		}
		positionbaks.Add(positions);
	}
	
	public void RestorePositions()
	{
		List<Position> oldpositions = positionbaks[positionbaks.Count-1];
		positionbaks.RemoveAt(positionbaks.Count-1);
			
		for (int i=0;i<entities.Count;i++)
		{
			Entity e = entities[i];
			Position oldpos = oldpositions[i];
			if (oldpos!=e.pos)
			{
				e.Translate(e.lastmovedir.Invert(),-1.0f);
				e.lastmovedir = Position.Direction.MAX;
			}
		}
		//CalcBeams();
	}
	public void DiscardPositions()
	{
		positionbaks.RemoveAt(positionbaks.Count-1);
	}
	
	public void PostMoveMessage(Entity e, Position.Direction d, Entity.MoverType othertype, float delay)
	{
		bool modified = true;
		while (modified)
		{
			modified = entities.Any( ent => ent.PostMoveMessage(e,d,othertype,delay) );
		}
	}
	
	public bool DoInput(Position.Direction d)
	{		
		PreMoveCalcs();
		
		entities.ForEach( e => e.PreMove() );
		Player p = Player();
		
		bool moved = p.TryMove(d,Entity.MoverType.Keyboard,0.0f);
		
		PostMoveCalcs();
		
		return moved;
	}
		
	
	public List<BeamDat> GetBeamsOverlapping(Position p)
	{
		return beams.Where( b => b.ContainsPos(p) ).ToList();
	}
	
	public IEnumerable<BeamDat> GetBeamsPushing(Position p)
	{
		return beams.Where( b => b.owner.pos+(b.length)*b.dir.ToPosition() == p );
	}
	
	
	public bool AutomaticTick()
	{
		PreMoveCalcs();
		
		bool result=false;
		List<Entity> elist = new List<Entity>(entities);		
		for (int i=0;i<entities.Count;i++)
		{
			Entity e = entities[i];
			if (e.beampushes)
			{
				IEnumerable<BeamDat> overlappingbeams = GetBeamsPushing(e.pos);
				
				if (e.facing!= Position.Direction.MAX)
				{
					overlappingbeams = overlappingbeams.Concat(GetBeamsPushing(e.pos+e.facing));
				}
				
				overlappingbeams = overlappingbeams.Where(b=>b.owner.lastmovedir== Position.Direction.MAX);
				if (overlappingbeams					
					.Select( b => b.dir )
					.Distinct()
					.Count()!=1)
					continue;
				
				BeamDat beam = overlappingbeams.First();
				
				entities.ForEach( ent => ent.PreMove() );		
				if (e.TryMove(beam.dir,Entity.MoverType.Beam,0.0f))
				{
					//Debug.Log(beam.owner.animdelay);
					result=true;
				}
			}			
		}
		
		PostMoveCalcs();
		
		//SanityCheck();
		return result;
	}	
	
	//EQUALITY BEGIN
	public override bool Equals(System.Object obj)
	{
		return obj is GameState && this == (GameState)obj;
	}
	
	public override int GetHashCode()
	{
		unchecked
		{			
			int hash=17;
			foreach(Entity e in entities)
			{
				hash = hash*29+e.GetHashCode();
			}
			return hash;
		}
	}
	
	public static bool operator ==(GameState x, GameState y)
	{		    
		// If both are null, or both are same instance, return true.
	    if (System.Object.ReferenceEquals(x, y))
	    {
	        return true;
	    }
	
	    // If one is null, but not both, return false.
	    if (((object)x == null) || ((object)y == null))
	    {
	        return false;
	    }
		// If one is null, but not both, return false.
	    if (((object)x == null) && ((object)y == null))
	    {
	        return true;
	    }	
		
		return Enumerable.SequenceEqual(x.entities,y.entities);		
	}
	
	public static bool operator !=(GameState x, GameState y)
	{
		return ! (x==y);
	}		
	//EQUALITY END
	
	//version 2 - added dir parameter
	public const int version = 4;
	
	public void Write(System.IO.Stream output)
	{
		System.IO.BinaryWriter bw = new System.IO.BinaryWriter(output);
		bw.Write((short)version);
		bw.Write(day);
		entities.ForEach( e => e.Write(bw) );
	}
	
	//need to callpostdeserialize after
	public static GameState Read(System.IO.Stream input)
	{
		System.IO.BinaryReader br = new System.IO.BinaryReader(input);
		int fileversion = br.ReadInt16();
		
		GameState gs = new GameState();
		
		
		if (fileversion>2)
		{
			gs.day = br.ReadBoolean();
		}
		else
		{
			gs.day=false;
		}
		
		
		while(br.PeekChar()>=0)
		{
			Entity e = Entity.Read(br,gs,fileversion);
			gs.entities.Add(e);
		}
		
		gs.PostDeserialize();
		return gs;
	}
	
	//need to callpostdeserialize after
	public GameState Clone()
	{
		GameState clone = new GameState();
		clone.day=day;
		foreach(Entity e in entities)
		{
			clone.entities.Add(e.Clone(clone));
		}
		clone.PostDeserialize();
		return clone;
	}
	
}
